//#define DEBUG_PRINT  // use to print out debugging data files
                     // comment out to silence

// JR001DLL.cpp and JR001DLL.h code by Jeff Vincent from
//   the RARS ROBOTDLL code, compiled as a DLL under
//   Bloodshed Dev++ C++ 4.01.
// Code started February 2004, then thrashed and submitted 
//   for RARS race 1, March 7, 2004 (still a long ways to go).
// Contact: jvincent@acmenet.net

// JR00102.cpp - 3/13/2004 for RARS race 2
//   converted to OOrobot by moderator Dennis L.
//   tweaked FUDGE_SPEED downward and added race2.trk - specific tweaks
//   added braking for collision avoidance

// JR00105.cpp - 4/3/2004 for RARS race 5
//   tweaked FUDGE_SPEED and added race5.trk - specific tweaks
 
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "os.h"
#include "car.h"
#include "track.h"

class JR001 : public Driver {
public:
   JR001::JR001()
   {
      m_sAuthor       = "Jeff Vincent";
      m_sDescription  = "?";
      m_iNoseColor    = oBLACK;
      m_iTailColor    = oGREEN;
      m_sModel3D      = NULL;
      m_sName         = "JR001";
      m_sBitmapName2D = "car_green_green";
   }

// This is the control function, for your robot
con_vec drive(situation& s)
{
    // Code for your robot goes here...
#ifdef DEBUG_PRINT
    FILE *fp1, *fp2, *fp3, *fp4;
    double vel, end, nxtvel;
    double a1, x, minrad, curve_offset, drt;
    double decel_vel, decel_rate, decel_dist;
#endif
    double lane, cf;

    static double width;

    static int init = 1;
    static int init2 = 11;

    const double BLIP = 0.20;
    const double TEST_SPEED = 200.0;

    int segcntr, nextseg, nextnextseg, btcntr;
    int sub_seg;
    static int lastseg, lastsub;
    double dlane, dprevlane, dvel, dprevvel, dstart, dend, dlen;
    double maxvel, nextmaxvel, prevmaxvel;
    static double elapsed_time = 0.0;
    double bd, bdlength;
   
    struct lane_data_struct
    {
        double lane;    // desired lane (distance from left edge of track)
        double vel;     // desired velocity
        double end;     // end of subsegment (distance from start of segment)
        int accel_flag; // how to accelerate/decelerate
                        //   0   means gradual accel/decel (average velocities)
                        //   1   means maximum accel/decel
        int prev_sub;   // previous subsegment code
                        //   -2  means last track segment (use lastseg, lastsub)
                        //   -1  means subtract one from current
                        // other means use other with segment-1
    };
    static struct lane_data_struct lane_data[MAX_SEGMENT+1][4];

    double current_path_angle, desired_path_angle, delta_path_angle;
    double desired_offset;
    double braking;

    const double DESIRED_RESTORE_TIME = 2.0; // was 3.0 -- goes freaky (excessive
                                             // oscillation) when <= 0.25 sec
    const double C1 = 1.95;  // was 1.95
    const double C2 = 1.95;  // was 1.95
    const double PCT_BRAKING = 100.0;
    const double FUDGE_SPEED = 1.20;     // was 1.20 for race1.trk, was 1.15 for race2.trk
                    // for race1.trk  - 1.05, 1.10 worked, 1.15 goes off on turn 8 hairpin
                    // with turn 8 tweak, 1.18 matches ALI2, works up to at least 1.3 (but more drifting)
                    // has to be lower than 1.18 for tracks with long turns (you eventually run out of room)                        
    const double CURVE_LANE = 5.0;
    const double PCT_OF_A = 0.99;
    const double MAX_ALPHA_SLOPE = log ( 1.0 - PCT_OF_A );  // -4.6
    
    const double MAX_ANGLE = 0.1;
    const double STEER_GAIN = 0.5;
    const double DAMP_GAIN = 2.0;
    const double BIG_SLIP = 14.0;
    const double LANE_ERROR = 2.0; // allowable deviation from lane before correction
                                   // eventually zero if I get a smooth algoritm

    if ( init != 0 )
    {

#ifdef DEBUG_PRINT
        fp1 = fopen ( "e:\\data\\rars\\rars\\vcdata.txt", "w" );
        fprintf ( fp1, "Delta_time = %f\n", delta_time );
        fprintf ( fp1, 
            "deltaPA  lane   s.to_lft  s.to_rgt  s.to_end   s.v       s.vn    angle   s.cen_a   s.tan_a    r.vc    r.alpha\n" );
        fprintf ( fp1, 
            "%5.2f   %6.2f   %6.2f    %7.2f   %6.2f   %6.2f   %6.3f   %6.2f    %6.2f   %6.2f   %6.3f\n",
            lane, s.to_lft, s.to_rgt, s.to_end, s.v, s.vn, sin(s.vn/s.v), s.cen_a, s.tan_a, result.vc, result.alpha );
        fclose ( fp1 );

        fp2 = fopen ( "e:\\data\\rars\\rars\\vcdata2.txt", "w" );
        for ( segcntr = 0; segcntr < currentTrack->NSEG; segcntr++ )
        {
            if ( currentTrack->lftwall[segcntr].radius == 0.0 )
            {
                fprintf ( fp2, "segment[%d].length = %f\n\n", segcntr, 
                            currentTrack->lftwall[segcntr].length );
            }
            else
            {
                fprintf ( fp2, "lftwall[%d].radius = %f\n", segcntr, 
                            currentTrack->lftwall[segcntr].radius );
                fprintf ( fp2, "rgtwall[%d].radius = %f\n", segcntr, 
                            currentTrack->rgtwall[segcntr].radius );
                fprintf ( fp2, "lftwall[%d].length = %f\n\n", segcntr, 
                            currentTrack->lftwall[segcntr].length );
            }
        }
        fclose ( fp2 );

        fp3 = fopen ( "e:\\data\\rars\\rars\\vcdata3.txt", "w" );
        fprintf ( fp3, "Time     Distance   Velocity\n");
        fclose ( fp3 );
#endif
      
        width = currentTrack->width;

        switch ( args.m_iSurface )
        {
            case 0 :
                cf = MYU_MAX0 * PCT_OF_A;
                break;
            case 1 :
                cf = MYU_MAX1 * PCT_OF_A;
                break;
            case 2 :
                cf = MYU_MAX2 * PCT_OF_A;
                break;
        }

        lastsub = -2;
        for ( segcntr = 0; segcntr < currentTrack->NSEG; segcntr++ )
        {
            nextseg = segcntr + 1;
            nextnextseg = nextseg + 1;
            if ( nextseg == currentTrack->NSEG )
            {
                nextseg = 0;
                nextnextseg = 1;
            }
            if ( nextnextseg == currentTrack->NSEG )
                nextnextseg = 0;

            if ( currentTrack->rgtwall[segcntr].radius == 0.0 )
                maxvel = 250.0;
            else
            {
                if ( currentTrack->rgtwall[segcntr].radius > 0.0 )
                {
                    maxvel = FUDGE_SPEED * sqrt ( g * cf * 
                        ( currentTrack->rgtwall[segcntr].radius - 
                          (width - CURVE_LANE) /* width / 2.0 */ ) );
                }
                else
                {
                    maxvel = FUDGE_SPEED * sqrt ( g * cf * 
                        ( -currentTrack->rgtwall[segcntr].radius + CURVE_LANE ) /* width / 2.0 */ );               
                }
            }            

            if ( currentTrack->rgtwall[nextseg].radius == 0.0 )
                nextmaxvel = 250.0;
            else
            {
                if ( currentTrack->rgtwall[nextseg].radius > 0.0 )
                {
                    nextmaxvel = FUDGE_SPEED * sqrt ( g * cf * 
                        ( currentTrack->rgtwall[nextseg].radius - 
                          (width - CURVE_LANE) /* width / 2.0 */ ) );
                }
                else
                {
                    nextmaxvel = FUDGE_SPEED * sqrt ( g * cf * 
                        ( -currentTrack->rgtwall[nextseg].radius + CURVE_LANE ) /* width / 2.0 */ );               
                }
            }

            if ( nextmaxvel < maxvel )  // must have accel + decel subsegments
            {                           //   either a straight into a turn
                bd = 0.0;               //   or a turn into a slower turn
                if ( segcntr != 0 )
                    prevmaxvel = lane_data[segcntr-1][lastsub].vel;
                else
                {
                    if ( currentTrack->rgtwall[currentTrack->NSEG-1].radius > 0.0 )
                    {
                        prevmaxvel = FUDGE_SPEED * sqrt ( g * cf * 
                            ( currentTrack->rgtwall[currentTrack->NSEG-1].radius - 
                              (width - CURVE_LANE) /* width / 2.0 */ ) );
                    }
                    else
                    {
                        prevmaxvel = FUDGE_SPEED * sqrt ( g * cf * 
                            ( -currentTrack->rgtwall[currentTrack->NSEG-1].radius 
                              + CURVE_LANE ) /* width / 2.0 */ );               
                    }
                }
                if ( currentTrack->rgtwall[segcntr].radius == 0.0 )
                    bdlength = currentTrack->rgtwall[segcntr].length;
                else
                    bdlength = fabs ( currentTrack->rgtwall[segcntr].length *
                        ( currentTrack->rgtwall[segcntr].radius + 
                          currentTrack->lftwall[segcntr].radius ) / 2.0 );                
#ifdef DEBUG_PRINT
    fp2 = fopen ( "e:\\data\\rars\\rars\\vcdata2.txt", "a" );
    fprintf ( fp2, "seg = %2d   bdlength = %f\n", segcntr, bdlength );
    fclose ( fp2 );
#endif
                if ( brake_test_vel 
                    ( prevmaxvel, bdlength, bd, cf ) > nextmaxvel )
                {
                    bd = bd + bdlength / 2.0;
                    for ( btcntr = 2; btcntr <= 12; btcntr++ )
                    {
                        if ( brake_test_vel 
                            ( prevmaxvel, bdlength, bd, cf ) > nextmaxvel )
                        {
                            bd = bd + bdlength / (1 << btcntr);
                        }
                        else
                        {
                            bd = bd - bdlength / (1 << btcntr);
                        }
                    }
                }
#ifdef DEBUG_PRINT
    fp2 = fopen ( "e:\\data\\rars\\rars\\vcdata2.txt", "a" );
    fprintf ( fp2, "seg = %2d   bd = %f\n", segcntr, bd );
    fclose ( fp2 );
#endif
/*
                if ( currentTrack->rgtwall[nextseg].radius > 0.0 )
                {
                    lane_data[segcntr][0].lane = CURVE_LANE; // width / 2.0;
                }
                else
                {
                    lane_data[segcntr][0].lane = width - CURVE_LANE; // width / 2.0;
                }
*/
/*
                if ( currentTrack->rgtwall[segcntr].radius == 0.0 )
                {
                    if ( currentTrack->rgtwall[nextseg].radius > 0.0 )
                    {
                        lane_data[segcntr][0].lane = CURVE_LANE;
                    }
                    else
                    {
                        lane_data[segcntr][0].lane = width - CURVE_LANE;
                    }
                }
                else
                {
                    if ( ( currentTrack->rgtwall[segcntr].radius / 
                           currentTrack->rgtwall[nextseg].radius ) < 0.0 )
                    {   // turn radii are opposite signs, so L-R or R-L esses
                        lane_data[segcntr][0].lane = width / 2.0;
                    }   // go towards middle of track
                    else    // turns are in same direction
                    {       // go towards that same side
                        if ( currentTrack->rgtwall[nextseg].radius > 0.0 )
                        {
                            lane_data[segcntr][0].lane = CURVE_LANE;
                        }
                        else
                        {
                            lane_data[segcntr][0].lane = width - CURVE_LANE;
                        }
                    }
                }
*/
                if ( currentTrack->rgtwall[segcntr].radius == 0.0 ) // str to...
                {
                    if ( currentTrack->rgtwall[nextseg].radius > 0.0 ) // ...left
                    {
                        lane_data[segcntr][0].lane = CURVE_LANE; // go far left
                    }
                    else 
                    {
                        if ( currentTrack->rgtwall[nextseg].radius < 0.0 ) // ...right
                        {
                            lane_data[segcntr][0].lane = width - CURVE_LANE; // go far right
                        }
                        else                        // ...another straight (possible?)
                        {
                            lane_data[segcntr][0].lane = width / 2.0; // go middle
                        }
                    }
                }
                else                                            // turn to...
                {
                    if ( currentTrack->rgtwall[nextseg].radius == 0.0 ) // ...str
                    {
                        if ( currentTrack->rgtwall[segcntr].radius > 0.0 ) // current left
                        {
                            lane_data[segcntr][0].lane = CURVE_LANE; // stay left
                        }
                        else                                            // current right
                        {
                            lane_data[segcntr][0].lane = width - CURVE_LANE; // stay right
                        }
                    }
                    else                                        // ...another turn
                    {
                        if ( ( currentTrack->rgtwall[segcntr].radius /
                               currentTrack->rgtwall[nextseg].radius ) < 0.0 )  // opposite direction
                        {   // turn radii are opposite signs, so L-R or R-L esses
                            lane_data[segcntr][0].lane = width / 2.0;   // go middle
                        }
                        else                                    // same direction
                        {
                            if ( currentTrack->rgtwall[segcntr].radius > 0.0 )  // current left
                            {
                                lane_data[segcntr][0].lane = CURVE_LANE; // stay left
                            }
                            else                                            // current right
                            {
                                lane_data[segcntr][0].lane = width - CURVE_LANE; // stay right
                            }
                        }
                        
                    }
                }

                lane_data[segcntr][0].vel = maxvel;
                if ( currentTrack->rgtwall[segcntr].radius == 0.0 )
                {
                    lane_data[segcntr][0].end = 
                        currentTrack->rgtwall[segcntr].length - bd;
                }
                else
                {
                    lane_data[segcntr][0].end = 
                        currentTrack->rgtwall[segcntr].length - bd / bdlength;
                }
                lane_data[segcntr][0].accel_flag = 1;
                lane_data[segcntr][0].prev_sub = lastsub;

                lane_data[segcntr][1].lane = lane_data[segcntr][0].lane;
                lane_data[segcntr][1].vel = nextmaxvel;
                lane_data[segcntr][1].end = 
                    currentTrack->rgtwall[segcntr].length;
                lane_data[segcntr][1].accel_flag = 1;
                lane_data[segcntr][1].prev_sub = -1;
                lastsub = 1;
            }
            else            // just one segment (a turn into a straight or
                            //                   a turn into a faster turn or
                            //                   a straight into a faster turn)
            {
/*
                if ( currentTrack->rgtwall[nextseg].radius != 0.0 )
                {
                    if ( ( currentTrack->rgtwall[segcntr].radius / 
                           currentTrack->rgtwall[nextseg].radius ) < 0.0 )
                    {   // turn radii are opposite signs, so L-R or R-L esses
                        lane_data[segcntr][0].lane = width / 2.0;
                    }   // go towards middle of track
                    else    // turns are in same direction
                    {       // go towards that same side
                        if ( currentTrack->rgtwall[nextseg].radius > 0.0 )
                        {
                            lane_data[segcntr][0].lane = CURVE_LANE; // width / 2.0;
                        }
                        else
                        {
                            lane_data[segcntr][0].lane = width - CURVE_LANE; // width / 2.0;
                        }
                    }
                }
                else
                {
                    if ( currentTrack->rgtwall[nextnextseg].radius > 0.0 )
                    {
                        lane_data[segcntr][0].lane = CURVE_LANE; // width / 2.0;
                    }
                    else
                    {
                        lane_data[segcntr][0].lane = width - CURVE_LANE; // width / 2.0;
                    }
                }
*/
                if ( currentTrack->rgtwall[segcntr].radius == 0.0 ) // str to...
                {
                    if ( currentTrack->rgtwall[nextseg].radius > 0.0 ) // ...left
                    {
                        lane_data[segcntr][0].lane = CURVE_LANE; // go far left
                    }
                    else 
                    {
                        if ( currentTrack->rgtwall[nextseg].radius < 0.0 ) // ...right
                        {
                            lane_data[segcntr][0].lane = width - CURVE_LANE; // go far right
                        }
                        else                        // ...another straight (possible?)
                        {
                            lane_data[segcntr][0].lane = width / 2.0; // go middle
                        }
                    }
                }
                else                                            // turn to...
                {
                    if ( currentTrack->rgtwall[nextseg].radius == 0.0 ) // ...str
                    {
                        if ( currentTrack->rgtwall[segcntr].radius > 0.0 ) // current left
                        {
                            lane_data[segcntr][0].lane = CURVE_LANE; // stay left
                        }
                        else                                            // current right
                        {
                            lane_data[segcntr][0].lane = width - CURVE_LANE; // stay right
                        }
                    }
                    else                                        // ...another turn
                    {
                        if ( ( currentTrack->rgtwall[segcntr].radius /
                               currentTrack->rgtwall[nextseg].radius ) < 0.0 )  // opposite direction
                        {   // turn radii are opposite signs, so L-R or R-L esses
                            lane_data[segcntr][0].lane = width / 2.0;   // go middle
                        }
                        else                                    // same direction
                        {
                            if ( currentTrack->rgtwall[segcntr].radius > 0.0 )  // current left
                            {
                                lane_data[segcntr][0].lane = CURVE_LANE; // stay left
                            }
                            else                                            // current right
                            {
                                lane_data[segcntr][0].lane = width - CURVE_LANE; // stay right
                            }
                        }
                        
                    }
                }

                lane_data[segcntr][0].vel = maxvel;
                lane_data[segcntr][0].end = currentTrack->rgtwall[segcntr].length;
                lane_data[segcntr][0].accel_flag = 0;   // TEST
                lane_data[segcntr][0].prev_sub = lastsub;
                lastsub = 0;
            }            


        }
        lastseg = currentTrack->NSEG - 1;

//seg = 6   subseg = 0   lane =  5.00   vel = 114.68   end =    1.08
//
//seg = 7   subseg = 0   lane = 81.00   vel = 250.00   end =    0.01
//seg = 7   subseg = 1   lane = 81.00   vel =  56.29   end =   51.35
//
// tweaking for braking approaching hairpin turn 8
// add a braking subsegment to segment 6 for track "race1.trk"
//
// also some added speed for turns 24, 25, and 27

        if ( strcmp ( currentTrack->m_sFileName, "race1.trk" ) == 0 )
        {
            lane_data[6][1].lane = lane_data[7][1].lane;    // add new subsegment
            lane_data[6][1].vel = lane_data[7][1].vel;
            lane_data[6][1].end = lane_data[6][0].end;
            lane_data[6][1].accel_flag = 1;
            lane_data[6][1].prev_sub = -1;
            lane_data[6][0].end = lane_data[6][1].end * 0.70;
                                            // brakepoint established empirically
            lane_data[7][0].prev_sub = 1;   // link next subsegment

            lane_data[24][0].vel *= 1.30;   // speed up fast turns
            lane_data[25][0].vel *= 1.30;
            lane_data[26][1].vel *= 1.20;
            lane_data[27][0].lane = width / 2.0;    // loosen up final turn exit
            lane_data[27][0].vel *= 1.20;
        }

//seg = 24   subseg = 0   lane = 36.00   vel = 110.80   end =    0.96
//seg = 24   subseg = 1   lane = 36.00   vel =  99.86   end =    1.24
//
//seg = 25   subseg = 0   lane =  5.00   vel =  99.86   end =    1.94
//
//seg = 26   subseg = 0   lane =  5.00   vel = 250.00   end =  582.89
//seg = 26   subseg = 1   lane =  5.00   vel = 125.71   end =  660.81
//
//seg = 27   subseg = 0   lane =  5.00   vel = 125.71   end =    1.20
//
// Race2.trk tweak

        if ( strcmp ( currentTrack->m_sFileName, "race2.trk" ) == 0 )
        {
            lane_data[24][1].vel *= 0.94;   // slow down long, fast turns
            lane_data[25][0].vel *= 0.94;
            lane_data[26][1].vel *= 0.97;
            lane_data[27][0].vel *= 0.97;
        }
        
// Race5.trk tweak
        if ( strcmp ( currentTrack->m_sFileName, "race5.trk" ) == 0 )
        {
            lane_data[1][0].vel *= 0.95;   // slow down long turns
            lane_data[7][0].vel *= 0.90;

            lane_data[14][0].end = lane_data[14][0].end * 0.90;
                                            // brakepoint established empirically
            lane_data[14][1].vel *= 0.90;
            lane_data[15][0].vel *= 0.90;
        }

/*
        lane_data[0][0].lane = s.to_lft;
        lane_data[0][1].lane = s.to_lft;
*/

#ifdef DEBUG_PRINT
        fp2 = fopen ( "e:\\data\\rars\\rars\\vcdata2.txt", "a" );
        for ( segcntr = 0; segcntr < currentTrack->NSEG; segcntr++ )
        {
            fprintf ( fp2, "seg = %d   subseg = %d   lane = %5.2f   vel = %6.2f   end = %7.2f\n", 
                segcntr, 0, lane_data[segcntr][0].lane, lane_data[segcntr][0].vel, lane_data[segcntr][0].end );
            if ( lane_data[segcntr][0].end != currentTrack->rgtwall[segcntr].length )
            {
                fprintf ( fp2, "seg = %d   subseg = %d   lane = %5.2f   vel = %6.2f   end = %7.2f\n", 
                    segcntr, 1, lane_data[segcntr][1].lane, lane_data[segcntr][1].vel, lane_data[segcntr][1].end );
            }
            fprintf ( fp2, "\n" );
        }
        fclose ( fp2 );                  
#endif

        init = 0;
    }
// END init


//   lane = ( s.to_lft + s.to_rgt ) * 0.1; // 0.5; // middle of the road

/*
    if ( ( ( s.to_lft - lane ) > -LANE_ERROR ) && ( s.vn < 2 ) )
    {
        result.alpha = 0.01;
    }
    else
    {
        if ( ( ( s.to_lft - lane ) < -LANE_ERROR ) && ( s.vn > -2 ) )
        {
            result.alpha = -0.01;
        }
        else
        {
            result.alpha = 0.0;
        }
    }

    if ( s.to_end <= 500.0 )
    {
        result.alpha = 0.0;
    }

    if ( ( init2 ) && ( s.to_end <= 400.0 ) && ( s.cur_rad == 0 ) )
    {
        result.alpha = BLIP;
        init2--;
        fp2 = fopen ( "e:\\data\\rars\\rars\\vcdata2.txt", "a" );
        fprintf ( fp2, "%f\n", s.vn );
        fclose ( fp2 );
    }
*/

    sub_seg = 0;        // find proper subsegment
    while ( ( s.cur_len - s.to_end ) > lane_data[s.seg_ID][sub_seg].end )
    {
        sub_seg++;
    }

    dlane = lane_data[s.seg_ID][sub_seg].lane;  // desired lane
    dvel = lane_data[s.seg_ID][sub_seg].vel;    // desired velocity
    dend = lane_data[s.seg_ID][sub_seg].end;    // desired end

    if ( lane_data[s.seg_ID][sub_seg].prev_sub == -1 )
    {
        dprevlane = lane_data[s.seg_ID][sub_seg-1].lane;    // desired previous lane
        dprevvel = lane_data[s.seg_ID][sub_seg-1].vel;      // desired previous velocity
        dlen = dend - lane_data[s.seg_ID][sub_seg-1].end;   // desired length
    }
    else
    {
        if ( lane_data[s.seg_ID][sub_seg].prev_sub == -2 )
        {
            dprevlane = lane_data[lastseg][lastsub].lane;
            dprevvel = lane_data[lastseg][lastsub].vel;
            dlen = dend;
        }
        else
        {
            dprevlane = lane_data[s.seg_ID-1]
                            [lane_data[s.seg_ID][sub_seg].prev_sub].lane;
            dprevvel = lane_data[s.seg_ID-1]
                            [lane_data[s.seg_ID][sub_seg].prev_sub].vel;
            dlen = dend;
        }
    }

    dstart = dend - dlen;   // desired start

    if ( dlane == dprevlane )
    {
        lane = dlane;
    }
    else
    {
//      lane = L0 + ( L1 - L0 ) * ( ( s.cur_len - s.to_end ) - sta0 ) / len
        lane = dprevlane + ( dlane - dprevlane ) *
               ( ( s.cur_len - s.to_end ) - dstart ) / dlen;
    }

//  dprevvel    vel = lane_data[s.seg_ID][sub_seg].vel0;
//  dend        end = lane_data[s.seg_ID][sub_seg].sta1;
//  dvel        nxtvel = lane_data[s.seg_ID][sub_seg].vel1;

/*
    sub_seg++;
    nxtvel = lane_data[s.seg_ID][sub_seg].vel;
    if ( nxtvel == 0.0 )
        nxtvel = lane_data[s.seg_ID+1][0].vel;
    if ( nxtvel == 0.0 )
        nxtvel = lane_data[0][0].vel;
*/

    current_path_angle = sin ( s.vn / s.v );
    desired_offset = s.to_lft - lane;

/*
    if ( s.cur_rad != 0.0 )
    {
        curve_offset = s.cur_rad - sqrt ( s.cur_rad * s.cur_rad -
                       s.v * s.v * DESIRED_RESTORE_TIME * DESIRED_RESTORE_TIME );
        desired_offset = desired_offset + curve_offset;
    }
    else
        if ( s.to_end <= 100.0 )
        {
            curve_offset = s.nex_rad - sqrt ( s.nex_rad * s.nex_rad -
                           s.v * s.v * DESIRED_RESTORE_TIME * DESIRED_RESTORE_TIME );
            desired_offset = desired_offset + curve_offset;
        }
*/
/*
    drt = DESIRED_RESTORE_TIME;
    if ( s.cur_rad == 0.0 ) drt *= 5.0;
*/

    desired_path_angle = sin ( ( desired_offset ) /
                                 ( s.v * DESIRED_RESTORE_TIME ) );
    delta_path_angle = desired_path_angle - current_path_angle;

#ifdef DEBUG_PRINT
    fp4 = fopen ( "e:\\data\\rars\\rars\\vcdata4.txt", "w" );
    fprintf ( fp4, "dead_ahead = %d\n", s.dead_ahead );
    for ( int i = 0; i < NEARBY_CARS; i++ )
    {
        fprintf ( fp4, "who = %2d   x = %5.2f   y = %5.2f   vx = %5.2f   vy = %5.2f\n", 
            s.nearby[i].who, s.nearby[i].rel_x, s.nearby[i].rel_y, 
            s.nearby[i].rel_xdot, s.nearby[i].rel_ydot );
    }
    fclose ( fp4 );
#endif

    braking = 0.0;
    for ( int i = 0; i < NEARBY_CARS; i++ )
    {
        if ( s.nearby[i].who < MAX_CARS )  // if there is a close car 
        {
            double x, y, vx, vy, d, v;
            
            x = s.nearby[i].rel_x;         // get forward distance 
            y = s.nearby[i].rel_y;         // get right distance 
            vx = s.nearby[i].rel_xdot;     // get forward relative speed 
            vy = s.nearby[i].rel_ydot;     // get lateral relative speed 
            d = sqrt( x*x + y*y );         // compute car-to-car distance 
            v = sqrt( vx*vx + vy*vy );     // compute separation velocity 
            if ( vy >= 0.0 )               // no action if cars are moving apart: 
                continue; 
            else if ( x * vx > 0.0 )       // no action if cars are moving apart: 
                continue; 
            else if ( d > 2.0 * CARLEN )   // no action if car is this far away: 
                continue; 
            // If the execution gets this far, it means that there is a car 
            // ahead of you, and getting closer, and less than two car lengths 
            //away.  Evaluate the situation more carefully to decide if 
            //evasive action is warranted. 
            
            braking = -5.0;
                        
/*
            if ( delta_path_angle >= 0 )
            {
                lane = lane + 10.0;
                if ( lane > width )
                    lane = width;
            }
            else
            {
                lane = lane - 10.0;
                if ( lane < 0.0 )
                    lane = 0.0;
            }
            desired_offset = s.to_lft - lane;
            desired_path_angle = sin ( ( desired_offset ) /
                                         ( s.v * DESIRED_RESTORE_TIME ) );
            delta_path_angle = desired_path_angle - current_path_angle;
*/
        }
    } 

#ifdef DEBUG_PRINT
    fp4 = fopen ( "e:\\data\\rars\\rars\\vcdata4.txt", "a" );
    fprintf ( fp4, "braking = %f\n", braking );
    fclose ( fp4 );

    if ( braking != 0.0 )
    {
        fp1 = fopen ( "e:\\data\\rars\\rars\\vcdata.txt", "a" );
        fprintf ( fp1, "braking!" );
        fclose ( fp1 );
    }
#endif

/*
    if ( s.dead_ahead == true )
    {
        if ( delta_path_angle >= 0 )
        {
            lane = lane + 10.0;
            if ( lane > width )
                lane = width;
        }
        else
        {
            lane = lane - 10.0;
            if ( lane < 0.0 )
                lane = 0.0;
        }
        desired_offset = s.to_lft - lane;
        desired_path_angle = sin ( ( desired_offset ) /
                                     ( s.v * DESIRED_RESTORE_TIME ) );
        delta_path_angle = desired_path_angle - current_path_angle;
    }
*/

#ifdef DEBUG_PRINT
    fp1 = fopen ( "e:\\data\\rars\\rars\\vcdata.txt", "a" );
    fprintf ( fp1, "%6.3f", delta_path_angle );
    fclose ( fp1 );
#endif

    result.alpha = delta_path_angle;

/*
    if ( ( fabs ( sin ( delta_path_angle ) ) * s.v / C2 ) > PCT_OF_A ) // was 1.0 , limit of 0.95 matches result of -3x
                                                                     // changed to 0.99 / -4.6x
    {
        result.alpha = ( -delta_path_angle / fabs ( delta_path_angle ) ) * ( C2 / s.v ) * MAX_ALPHA_SLOPE;

        fp1 = fopen ( "e:\\data\\rars\\rars\\vcdata.txt", "a" );
        fprintf ( fp1, "*  " );
        fclose ( fp1 );
    }
    else
    {
        result.alpha = ( -delta_path_angle / fabs ( delta_path_angle ) ) * ( C2 / s.v ) *
                       log ( 1 - ( fabs ( delta_path_angle ) * s.v / C2 ) );
        fp1 = fopen ( "e:\\data\\rars\\rars\\vcdata.txt", "a" );
        fprintf ( fp1, "   " );
        fclose ( fp1 );
    }
*/

// find desired alpha to achieved delta_path)angle (experimental data + friction function)
// if alpha is too high for given s.v, reduce

//    if ( s.cur_rad == 0.0 ) result.alpha = result.alpha / 12.0;

/*    result.alpha = STEER_GAIN * (s.to_lft - lane) - DAMP_GAIN * s.vn; */      // steering

// gas / brake

/*
    if ( s.to_end >= 1000 )
        result.vc = 250.0;
    else
    {
        result.vc = result.vc * ( 100.0 - PCT_BRAKING ) / 100.0;
        fp2 = fopen ( "e:\\data\\rars\\rars\\vcdata2.txt", "a" );
        fprintf ( fp2, "%7.2f   %6.2f   %6.2f\n", s.to_end, s.v, s.tan_a );
        fclose ( fp2 );
    }
*/

//       double lane0, lane1;
//       double vel0, vel1;
//       double sta0, sta1, len;
//       int accel_flag;

    if ( dprevvel == dvel )
    {
        result.vc = dvel; // was s.v + 5.0;
    }
    else // dprevvel != dvel
    {
        if ( lane_data[s.seg_ID][sub_seg].accel_flag == 0 ) 
                            // interpolate velocity from start/end velocities
        {
//          vc = v0 + ( v1 - v0 ) * ( ( s.cur_len - s.to_end ) - sta0 ) / len
            result.vc = dprevvel + ( dvel - dprevvel ) *
                ( ( s.cur_len - s.to_end ) - dstart ) / dlen;
        }
        else // full acceleration / deceleration
        {
            if ( dprevvel >= dvel ) // full braking
            {
                result.vc = dvel;

/*
        decel_vel = ( s.v + nxtvel ) / 2.0;
        decel_rate = g * 1.05 + ( decel_vel * decel_vel / 11400.0 );
        decel_dist = decel_vel * ( s.v - nxtvel ) / decel_rate;
        if ( s.to_end - end > decel_dist )
            result.vc = vel; // s.v + 5.0;
        else
//          result.vc = nxtvel;
            result.vc = s.v - 2.5;

//  The lane algorithm now allows you to select a lane at various points
//  on a given segment, but the high-G braking causes the car to skid and
//  doesn't allow tunring/lane changing; the line above uses lower-G
//  braking and allows the car to follow the lane (but invalidates my
//  braking assumtions and I miss my turns now. :)
*/

            }
            else // full acceleration
            {
                result.vc = dvel;
            }
        }
    }

    if ( braking < 0.0 )
    {
        result.vc = s.v + braking;
    }

    result.request_pit   = 0;           // must be 1 if you want to pit
    result.fuel_amount   = MAX_FUEL;    // fuel you want added at start/pitstop
    result.repair_amount = 0;           // amount of damage to repair at pitstop

// pitstop just after halfway if fuel stop will be necessary
    if ( ( ( s.laps_done == s.laps_to_go ) || ( s.laps_done+1 == s.laps_to_go ) )
            && ( s.fuel < MAX_FUEL / 2 ) )
    {
        result.request_pit = 1;
        result.fuel_amount = MAX_FUEL;
        if ( s.damage >= 20000 )
            result.repair_amount = 5000;
        else
            result.repair_amount = 0;
    }
    else
    {    
    // pitstop if low on fuel (may not make another lap)
        if ( s.fuel <= 3.0 )
        {
            result.request_pit = 1;
            result.fuel_amount = MAX_FUEL;
            if ( s.damage >= 25000 )
                result.repair_amount = 5000;
            else
                result.repair_amount = 0;
        }
        else
        {
        // pitstop if damage is excessive
            if ( s.damage >= 25000 )
            {
                result.request_pit = 1;
                result.repair_amount = 5000;
                result.fuel_amount = MAX_FUEL;
            }
        }
    }

#ifdef DEBUG_PRINT
    fp1 = fopen ( "e:\\data\\rars\\rars\\vcdata.txt", "a" );
    fprintf ( fp1, 
        "%5.2f   %6.2f   %6.2f    %7.2f   %6.2f   %6.2f   %6.3f   %6.2f    %6.2f   %6.2f   %6.3f\n",
        lane, s.to_lft, s.to_rgt, s.to_end, s.v, s.vn, sin(s.vn/s.v), s.cen_a, s.tan_a, result.vc, result.alpha );
    fclose ( fp1 );

    fp3 = fopen ( "e:\\data\\rars\\rars\\vcdata3.txt", "a" );
    fprintf ( fp3, "%6.3f   %7.2f   %6.2f\n",
        elapsed_time, s.cur_len - s.to_end, s.v );
    fclose ( fp3 );

    elapsed_time += delta_time;
#endif
   
    return result;
}

// Improvements:
// * create lane_data info for use on start/finish straight on first lap only
// * when steering on straights, try to steer only on non-braking segments
// * optimize turns by calculating larger radius with wider entry/exit
//     this may mean another segment on entry/exit straights
// * need to calculate braking point for turns entering tighter turns as well
//     (not just straights)
// 

private:
   con_vec result;  // use this to return your control parameters

// Determines velocity over a given acceleration/braking distance, 
//     to test for latest braking point
double brake_test_vel ( double v0, double s, double bd, double cf )
{
    double dist = 0.0;
    double vel = v0;
    double a, dvel;
/*
    fp = fopen ( "brake.txt", "a" );
    fprintf ( fp, "bd = %7.2f\n", bd );
    fprintf ( fp, "distance   velocity   accel\n" );
    fclose ( fp );
*/
    while ( dist < s - bd ) // accelerating
    {
        a = g * cf * ( 0.0000148 * vel * vel - 0.00755 * vel + 1 );
        dvel = a * delta_time;
        dist = dist + ( vel + dvel / 2.0 ) * delta_time;
        vel = vel + dvel;
/*        
        fp = fopen ( "brake.txt", "a" );
        fprintf ( fp, "%7.2f   %6.2f   %6.2f\n", dist, vel, a );
        fclose ( fp );
*/
    }
    
    while ( dist < s ) // braking
    {
        a = -g * cf * ( 0.00000266 * vel * vel + 0.00000887 * vel + 1 );
        dvel = a * delta_time;
        dist = dist + ( vel + dvel / 2.0 ) * delta_time;
        vel = vel + dvel;
/*        
        fp = fopen ( "brake.txt", "a" );
        fprintf ( fp, "%7.2f   %6.2f   %6.2f\n", dist, vel, a );
        fclose ( fp );
*/        
        if ( vel <= 0.0 )
        {
            return 0.0;
        }
    }
/*
    fp = fopen ( "brake.txt", "a" );
    fprintf ( fp, "vel returned = %6.2f\n\n", vel );
    fclose ( fp );
*/

    return vel;
}

};

Driver * getJR001Instance()
{ 
   return new JR001();
}
