/*======================================================================================
	Alison Chorley 
 	alison@csc.liv.ac.uk
	07/03/2003
	
	Filename:  ali2.cpp
  	Robot:     Ali2
  	Races:     all
  	Source:    Public
  	Color:     Blue/Blue
  	Data:      no need for a data file.

	Version 2.2
		Small changes to the driving style
		Change to the fuel tank to take MAX_FUEL at start and when pitting.
=====================================================================================*/

#include "car.h"

//--------------------------------------------------------------------------------------
//			PROTOTYPES
//--------------------------------------------------------------------------------------
con_vec drive(situation &s);
double corneringSpeed(double radius);
double brakingDistance(double speedNow, double speedNext, double brake_a);
void driveStraight(situation &s);
void driveCorner(situation &s);
void passing(situation &s);

//======================================================================================
//			CLASS Ali2
//======================================================================================

class Ali2 : public Driver
	{
	public:

//--------Class Variables---------------------------------------------------------------
		double alpha;
		double vc;

//--------Constructor-------------------------------------------------------------------
		Ali2::Ali2()
			{
			m_sName = "Ali2";
			m_sAuthor = "Alison Chorley";
			m_iNoseColor = oBLUE;
			m_iTailColor = oBLUE;
			m_sBitmapName2D = "car_blue_blue";
			}
//--------End of Constructor------------------------------------------------------------

//======================================================================================
//			Functions
//======================================================================================

//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
//----------MAIN DRIVE FUNCTION---------------------------------------------------------
//\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/

	con_vec drive(situation &s)
		{
		con_vec result;

		if(s.starting)
			{
			result.fuel_amount = MAX_FUEL;     // fuel when starting
			}

//--------------------------------------------------------------------------------------
//Simulator provided
		if(stuck( s.backward, s.v, s.vn, s.to_lft, s.to_rgt, &result.alpha, &result.vc))
			{
			return result;
			}

//--------------------------------------------------------------------------------------
//decide which function to call
		if (s.cur_rad == 0)
//drive straight
			driveStraight(s);
		else
			driveCorner(s);
//--------------------------------------------------------------------------------------
//called if there is a car nearby
		if (s.nearby[0].who < 16)
			passing(s);

//--------------------------------------------------------------------------------------
//decide to pit
		if ((s.fuel < 10)|| (s.damage > 20000))
			{
			result.request_pit = 1;
			result.repair_amount = s.damage;
			result.fuel_amount = MAX_FUEL;
			}
		else
			result.request_pit = 0;

//--------------------------------------------------------------------------------------
//return results to sim		
		result.vc =vc;
		result.alpha = alpha;

		return result;
		}

//---------End of Main Drive function---------------------------------------------------

//----------CORNERING SPEED-------------------------------------------------------------
	double corneringSpeed(double radius)
		{
		double rad;
		double coef_friction = 0.95;	//estimate of coefficient of friction
		double nextSpeed;

		if(radius < 0)
			rad = -radius;
		else
			rad = radius;

		nextSpeed = sqrt(rad * 32.2 * coef_friction);
		return (nextSpeed + 10);
		}
//--------End of Cornering Speed Function-----------------------------------------------

//----------BRAKING DISTANCE------------------------------------------------------------
	double brakingDistance(double speedNow, double speedNext, double brake_a)
		{
		double change;				//This is the total change in speed
		double distance;

		change = speedNext - speedNow;

		if(change > 0.0)			//need to accelerate not brake
			return(0);				//test here so don't need to do this in main function
		else
			{
			distance = change * (speedNow + 0.5 * change) / brake_a;
			return(distance);
			}
		}
//--------End of Braking Distance Function----------------------------------------------

//----------DRIVING STRAIGHT------------------------------------------------------------
	void driveStraight(situation &s)
		{
//--------------------------------------------------------------------------------------
//function variables
//lane widths, different depending whether you are going straight, left or right
		double width = (s.to_lft + s.to_rgt);
		double sLane = width / 2;			//middle of track for straights
		double lLane = 5;						//left side for left curves
		double rLane = width - 5;			//right side for right curves
//steering variables
		double steeringGain = 0.5;			//set for left turn
		double dampGain = 1.5;				//slide car if heading for wall
//braking variables
		double brake_a = -35;				//braking accelereation
//speed variables
		double nextSpeed = 0;				

//--------------------------------------------------------------------------------------
//calculating speed for next section
		if(s.nex_rad > 0)						//going left next
			nextSpeed = corneringSpeed(s.nex_rad + lLane);
		else if(s.nex_rad < 0)				//going right next
			nextSpeed = corneringSpeed(s.nex_rad -lLane);//because next radius is negative
		else										//straight
			nextSpeed = 500;					//fast as possible

//--------------------------------------------------------------------------------------
//Driving
		if(s.to_end > brakingDistance(s.v, nextSpeed, brake_a))	//outside braking zone
			{
			alpha = steeringGain * (s.to_lft - sLane)/width - dampGain*(s.vn/s.v);
			vc = 500;				//going fast
			}
		else
			{
//Calculate alpha
			if(s.nex_rad > 0)		//going left
				alpha = steeringGain *(s.to_lft - lLane)/width - dampGain*(s.vn/s.v);
			else						//going right
				alpha = steeringGain *(s.to_lft - rLane)/width - dampGain*(s.vn/s.v);
//Calculate speed
			if(s.v > nextSpeed) 
				vc = s.v + brake_a;		// brake
			else 
				vc = nextSpeed;				 
			}
		}
//--------End of Driving Straight Function----------------------------------------------

//----------DRIVE CORNERS---------------------------------------------------------------
	void driveCorner(situation &s)
		{
		double width = (s.to_lft + s.to_rgt);
//lane widths, different depending whether you are going straight, left or right
		double sLane = width / 2;		//middle of track for straights
		double lLane = 5;					//left side for left curves
		double rLane = width - 5;		//right side for right curves

		double nextSpeed = 0;			//initialise to zero just in case
		double steeringGain = 0.5;		//set for left turn
		double dampGain = 1.5;			//slide car if heading for wall

		double brake_a = -27;			//braking acceleration in a curve
		double brake = 4;					//reduce tire speed by this in a curve
		
		double currentSpeed = 0;
		double curveLength = 0;
		double curveToEnd = 0;

		if (s.cur_rad > 0)	//on left corner
			{
			currentSpeed = corneringSpeed(s.cur_rad + lLane);
			curveLength = s.cur_len*(s.cur_rad + lLane);
			curveToEnd = s.to_end*(s.cur_rad + lLane);
			}
		else			//on right corner
			{
			currentSpeed = corneringSpeed(s.cur_rad - lLane);
			curveLength = -s.cur_len*(s.cur_rad - lLane);
			curveToEnd = -s.to_end*(s.cur_rad - lLane);
			}

//--------------------------------------------------------------------------------------
//calculating speed for next section
		if(s.nex_rad > 0)				//going left
			nextSpeed = corneringSpeed(s.nex_rad + lLane);
		else if(s.nex_rad < 0)		//going right
			nextSpeed = corneringSpeed(s.nex_rad - lLane);	//cos next radius is negative
		else								//straight
			nextSpeed = 500;			//fast as possible

//--------------------------------------------------------------------------------------
//Driving
		if (currentSpeed > nextSpeed)		//slowing down
			{
			if(curveToEnd > brakingDistance(s.v, nextSpeed, brake_a))//outside braking zone
				{
//Calculate alpha
				if (s.cur_rad > 0)	//left
					alpha = steeringGain *(s.to_lft - lLane)/width - dampGain*(s.vn/s.v);
				else						//right
					alpha = steeringGain *(s.to_lft - rLane)/width - dampGain*(s.vn/s.v);
//calculate speed			
				vc = (currentSpeed) / cos(alpha);
				}
			else		//within braking distance
				{
//Calculate alpha
				if (s.nex_rad > 0)		//going left
					alpha = steeringGain *(s.to_lft - lLane)/width - dampGain*(s.vn/s.v);
				else							//going right
					alpha = steeringGain *(s.to_lft - rLane)/width - dampGain*(s.vn/s.v);
//Calculate speed	
				if (s.v > (nextSpeed + 4))	//going too fast
					vc = s.v - brake;
				else
					vc = nextSpeed/ cos(alpha);
				}
			}
		else		//accelerating
			{
			if (curveToEnd > curveLength/10)	//far to end
				{
//Calculate alpha
				if (s.cur_rad > 0)	//left
					alpha = steeringGain *(s.to_lft - lLane)/width - dampGain*(s.vn/s.v);
				else						//right
					alpha = steeringGain *(s.to_lft - rLane)/width - dampGain*(s.vn/s.v);
//calculate speed			
				vc = (currentSpeed) / cos(alpha);
				}
			else if (s.nex_rad == 0)	//going straight
				{
				alpha = steeringGain *(s.to_lft - sLane)/width - dampGain*(s.vn/s.v);
//increse speed
				vc = nextSpeed/cos(alpha);
				}
			else		//close to end
				{
//calcuate alpha
				if(s.nex_rad > 0)		//going left
					alpha = steeringGain *(s.to_lft - lLane)/width - dampGain*(s.vn/s.v);
				else
					alpha = steeringGain *(s.to_lft - rLane)/width - dampGain*(s.vn/s.v);
//calcuate speed
				vc = nextSpeed/cos(alpha);
				}
			}
		}
//--------End of Driving Corner Function------------------------------------------------

//----------PASSING CODE----------------------------------------------------------------
	void passing(situation &s)
		{
		double width = (s.to_lft + s.to_rgt);
//lane widths, different depending whether you are going straight, left or right
		double sLane = width / 2;		//middle of track for straights
		double lLane = 0;					//left side for left curves
		double rLane = width;		//right side for right curves

		double nextSpeed = 0;			//initialise to zero just in case
		double steeringGain = 0.9;		//0.5 set for left turn
		double dampGain = 1.5;			//2 slide car if heading for wall

		double brake_a = -27;			//braking acceleration in a curve
		double brake = 4;					//reduce tire speed by this in a curve
		
		double curveLength = 0;
		double curveToEnd = 0;

		double other_to_lft = s.to_lft + s.nearby[0].rel_x;
		double passLane = 0;

		if (s.cur_rad > 0)	//on left corner
			{
			curveLength = s.cur_len*(s.cur_rad + lLane);
			curveToEnd = s.to_end*(s.cur_rad + lLane);
			}
		else			//on right corner
			{
			curveLength = -s.cur_len*(s.cur_rad - lLane);
			curveToEnd = -s.to_end*(s.cur_rad - lLane);
			}

//calculating speed for next section
		if(s.nex_rad > 0)				//going left
			nextSpeed = corneringSpeed(s.nex_rad + lLane);
		else if(s.nex_rad < 0)		//going right
			nextSpeed = corneringSpeed(s.nex_rad - lLane);	//cos next radius is negative

//Calculate braking distance
		double brakingDist = brakingDistance(s.v, nextSpeed, brake_a);
		
//--------------------------------------------------------------------------------------
//Passing code
		if ((s.nearby[0].rel_ydot > 0)||(s.nearby[0].rel_y > (10 * CARLEN)))
			{}//nothing.  Moving apart
		else
			{
			if (s.cur_rad ==0)	//straight now
				{
				if ((other_to_lft >= (sLane - 1.5*CARWID))
					&&(other_to_lft <= (sLane + 1.5*CARWID)))
					{
					if (brakingDist < s.to_end)	//not in braking zone
						{
						if (s.nex_rad > 0)	//left next
							{
							passLane = other_to_lft - 1.5*CARWID;
							alpha = steeringGain*(s.to_lft-passLane)/width - dampGain*(s.vn/s.v);
							}
						else	//right next
							{
							passLane = other_to_lft + 1.5*CARWID;
							alpha = steeringGain*(s.to_lft-passLane)/width - dampGain*(s.vn/s.v);
							}
						}
					else	//inside braking zone
						{}//do nothing, let driveCorner() drive for the corner
					}
				else	//car to far to side to worry about
					{}//do nothing
				}
//--------------------------------------------------------------------------------------
			else if (s.cur_rad > 0)	//left now
				{
				if (s.nex_rad >= 0)	//left or straight next
					{
					if (other_to_lft > CARWID)		//far to the right
						{}//do nothing
					else
						{
						passLane = other_to_lft + 1.5*CARWID;
						alpha = steeringGain*(s.to_lft-passLane)/width - dampGain*(s.vn/s.v);
						}
					}
				else	//right next
					{
					if (other_to_lft >sLane)	//far to right
						{}//do nothing
					else if (curveToEnd > (curveLength / 2))//less than half way
						{
						if (other_to_lft > CARWID)
							{}//do nothing
						else
							{
							passLane = other_to_lft + 1.5*CARWID;
							alpha = steeringGain*(s.to_lft-passLane)/width - dampGain*(s.vn/s.v);
							}
						}
					else	//close to end
						{
						passLane = other_to_lft + 1.5*CARWID;
						alpha = steeringGain*(s.to_lft-passLane)/width - dampGain*(s.vn/s.v);
						}	
					}
				}
//--------------------------------------------------------------------------------------
			else	//right now
				{
				if (s.nex_rad <= 0)	//right or straight next
					{
					if (other_to_lft < (width - CARWID))		//far to the right
						{}//do nothing
					else
						{
						passLane = other_to_lft - 1.5*CARWID;
						alpha = steeringGain*(s.to_lft-passLane)/width - dampGain*(s.vn/s.v);
						}
					}
				else	//left next
					{
					if (other_to_lft < sLane)	//far to left
						{}//do nothing
					else if (curveToEnd > (curveLength / 2))//less than half way
						{
						if (other_to_lft < (width - CARWID))
							{}//do nothing
						else
							{
							passLane = other_to_lft - 1.5*CARWID;
							alpha = steeringGain*(s.to_lft-passLane)/width - dampGain*(s.vn/s.v);
							}
						}
					else	//close to end
						{
						passLane = other_to_lft - 1.5*CARWID;
						alpha = steeringGain*(s.to_lft-passLane)/width - dampGain*(s.vn/s.v);
						}	
					}
				}
			}
		}
//----------End of PASSING CODE-------------------------------------------------------
	};
//============================End of CLASS===============================================

Driver * getAli2Instance()
	{ 
	return new Ali2();
	}
//============================END========================================================
